///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoClient.cc
// -------------
// Cego client program
//
// Design and Implementation by Bjoern Lemke               
//                                                         
// (C)opyright 2000-2010 Bjoern Lemke                        
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: main
// 
// Description: 
//
// Status: QG-2.5.5
//
///////////////////////////////////////////////////////////////////////////////

#include <locale.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "Chain.h"
#include "Exception.h"
#include "GetLongOpt.h"
#include "File.h"
#include "Timer.h"
#include "Tokenizer.h"
#include "Net.h"
#include "SigHandler.h"

#include "Element.h"
#include "Document.h"
#include "XMLSuite.h"

#include "CegoOutput.h"
#include "CegoViewObject.h"
#include "CegoProcObject.h"
#include "CegoKeyObject.h"
#include "CegoDbHandler.h"
#include "CegoModule.h"

#include "CegoDefs.h"


#define USAGE "Usage: cgclt --user=<user>/<password>\n\
          [ --server=<host>]\n\
          [ --port=<port> ]\n\
	  [ --tableset=<tableset> ]\n\
          [ --batchfile=<batchfile> ]\n\
          [ --cmd=<command> ]\n\
          [ --logfile=<logfile> ]\n\
          [ --raw ]\n\
          [ --ignore ]\n\
          [ --debug ] [ --version  ] [ --help ]"


#define DEFAULTPORT 2200
#define DEFAULTSERVER "localhost"

#define HISTFILE ".cgclt_history"
#define PROMPT "CGCLT > "
#define PROMPT2 "> "


#define LANGUAGE_ENV "LANG"
#define LOCALE_CATEGORY LC_TIME

enum RunMode { INTERACTIVE, BATCH, COMMAND };

extern char __baseVersionString[];
extern char __xmlVersionString[];

bool flowCtl = false;
int flowCount = 10;

class AbortHandler : public SigHandler
{
  
public:

    AbortHandler(const Chain& serverName, 
		 int portNo,
		 const Chain& tableSet,
		 const Chain& user, 
		 const Chain& password,		 
		 long tid, 
		 CegoModule *pModule);
    ~AbortHandler();
    void sigCatch(int sig);

    void setQueryAbort(bool doAbort);
    
private:

    long _tid;
    bool _doAbort;
    CegoModule *_pModule;
    int _portNo;
    Chain _serverName;
    Chain _user;
    Chain _password;
    Chain _tableSet;
};

bool dispatchCmd(CegoDbHandler* pNetHandle, AbortHandler* pAH, Chain cmd, bool outMode, int & exitCode);

int main(int argc, char **argv)
{	
    
    GetLongOpt longOpt(argc, argv);
    
    longOpt.addOpt("version");
    longOpt.addOpt("help");
    longOpt.addOpt("logfile");
    longOpt.addOpt("tableset");
    longOpt.addOpt("user");
    longOpt.addOpt("cmd");
    longOpt.addOpt("batchfile");
    longOpt.addOpt("debug");
    longOpt.addOpt("raw");
    longOpt.addOpt("ignore");
    longOpt.addOpt("server", DEFAULTSERVER);
    longOpt.addOpt("port", Chain(DEFAULTPORT));
    longOpt.addOpt("simulate");
    
    try
    {
	longOpt.parseOpt(); 
    }
    catch ( Exception e )
    {
	Chain msg;
	e.pop(msg);
	cerr << msg << endl;
	cerr << USAGE << endl;
	exit(1);	
    }
    
    Chain user;
    Chain password;
    

#ifdef HAVE_MINGW32
    Chain homeDir(getenv("HOMEPATH"));
    Chain histPath = homeDir + "\\" + Chain(HISTFILE);
#else
    Chain homeDir(getenv("HOME"));
    Chain histPath = homeDir + "/" + Chain(HISTFILE);
#endif
    
    RunMode runMode = INTERACTIVE;

    bool debug=false;
    if ( longOpt.isSet("debug") )
	debug=true;

    bool rawMode = false;
    if ( longOpt.isSet("raw") )
	rawMode=true;

    bool ignoreError = false;
    if ( longOpt.isSet("ignore") )
	ignoreError=true;

    
    if ( longOpt.isSet( Chain("help") ) )
    {	
	cerr << USAGE << endl;
	exit(0);
    }

    if ( longOpt.isSet( Chain("version") ) )
    {
	cout << CEGO_PRODUCT << " (" << sizeof(long) * 8 << " bit), Version " << CEGO_VERSION 
	     << " [ Base : " << __baseVersionString  
	     << ", XML : " <<  __xmlVersionString << " ]" << endl;
	cout << CEGO_COPYRIGHT << endl;
	exit(0);
    }

    Chain logFile = longOpt.getOptValue("logfile");
    Chain serverName = longOpt.getOptValue("server");
    Chain tableSet = longOpt.getOptValue("tableset");
    int portNo = longOpt.getOptValue("port").asInteger();

    Chain batchFileName;
    Chain command;

    if ( longOpt.isSet("batchfile") )
    {
	runMode = BATCH;
	batchFileName = longOpt.getOptValue("batchfile");
    }
    else if ( longOpt.isSet("cmd") )
    {
	runMode = COMMAND;
	command = longOpt.getOptValue("cmd");
    }

    Chain authString = longOpt.getOptValue("user");
		
    Tokenizer authTok(authString, Chain("/")); 
    authTok.nextToken(user);
    authTok.nextToken(password);

    if ( user.length() == 0 )
    {
	cerr << "User not set" << endl;
	exit (1);
    }
    if ( password.length() == 0 )
    {
	cerr << "Password not set" << endl;
	exit (1);
    }
    if ( tableSet.length() == 0 )
    {
	cerr << "Tableset not set" << endl;
	exit (1);
    }

    // for localization
    char* lang = 0;
    if ( ( lang = getenv(LANGUAGE_ENV) ) != 0)
    {
	if ( setlocale(LOCALE_CATEGORY, lang) == 0)
	{
	    Chain msg = Chain("Cannot set locale ") + Chain(lang);
	    cerr << msg << endl;
	    exit(1);
	}
    }

    CegoModule *pModule;
    if ( logFile.length() > 0 )
	pModule = new CegoModule(logFile);
    else
	pModule = new CegoModule;

    
    unsigned long modId = pModule->getModId("CegoClient");
    
    if ( debug )
    {
	pModule->logModule(modId, "CegoClient", Logger::DEBUG);
	unsigned long id = pModule->getModId("CegoDbHandler");
	pModule->logModule(id, "CegoDbHandler", Logger::DEBUG);
    }
    else
    {
	pModule->logModule(modId, "CegoClient", Logger::NOTICE);
	unsigned long id = pModule->getModId("CegoDbHandler");
	pModule->logModule(id, "CegoDbHandler", Logger::NOTICE);
    }

    int exitCode;

    switch ( runMode )
    {

    case COMMAND:
    {
	
	Net n (  NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	NetHandler* pN;
	
	pN = n.connect(serverName, portNo);
	
	CegoDbHandler* pSH = new CegoDbHandler(pN, pModule);
	
	CegoDbHandler::ResultType res = pSH->requestSession(tableSet, user, password);
	if ( res != CegoDbHandler::DB_OK )
	{
	    Chain msg;
	    pSH->getMsg(msg);
	    throw Exception(EXLOC, msg);
	}
			
	dispatchCmd(pSH, 0, command, rawMode, exitCode);

	pSH->closeSession();
	
	delete pSH;
	delete pN;
	
	exit(exitCode);
			
	break;
    }
    case BATCH:
    {
	
	int lineNo=0;
	    
	try 
	{

	    Net n (  NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	    NetHandler* pN;
	    
	    pN = n.connect(serverName, portNo);

	    CegoDbHandler* pSH = new CegoDbHandler(pN, pModule);

	    CegoDbHandler::ResultType res = pSH->requestSession(tableSet, user, password);
	    if ( res != CegoDbHandler::DB_OK )
	    {
		Chain msg;
		pSH->getMsg(msg);
		throw Exception(EXLOC, msg);
	    }
	    
	    File cmdFile(batchFileName);
	    
	    cmdFile.open(File::READ);
	    
	    Chain cmd;
	    Chain line;
	    bool disableDelimiter=false;
	    bool goOn = true;

	    while (goOn && cmdFile.readLine(line))
	    {
		lineNo++;
		line = line.cutTrailing(" \t");

		bool isCommentLine = false;
		
		if ( line.length() >= 2 )
		{		    
		    if (line.cutTrailing(" \t").subChain(1, 2) == Chain("--"))
		    {
			isCommentLine = true;
		    }
		}
		
		if ( ! isCommentLine )
		{

		    if ( line == Chain("@") )
		    {
			if ( disableDelimiter == false )
			    disableDelimiter=true;
			else
			    disableDelimiter=false;
		    }
		    else
		    {
			// cmd = line;
			cmd = cmd + Chain(" ") + line;
			cmd = cmd.cutTrailing(" \t");
		    }
			 
		    if ( cmd.length() > 0 )
		    {

			if (cmd.subChain(cmd.length()-1, cmd.length()) == Chain(";") 
			    && disableDelimiter==false)
			{
			    			
			    // send command to server here
				
			    goOn = dispatchCmd(pSH, 0, cmd, rawMode, exitCode);
			    
			    if ( exitCode != 0 && ignoreError )
				goOn = true;

			    cout << "ok" << endl;			    			    			    
			    
			    cmd = Chain("");
			    
			}
		    }
		}
	    }
	    if ( cmd.length() > 1 )
	    {
		cout << "incomplete command ..<" << cmd << ">" << endl;
	    }
	    
	    cmdFile.close();

	    pSH->closeSession();

	    delete pSH;
	    delete pN;

	}
	catch ( Exception e)
	{
	    cout << "Line Number " + Chain(lineNo) + " : " << endl;
	    e.print();
	    exitCode=1;
	}
	break;
    }
    case INTERACTIVE:
    {
	try 
	{	    	    


	    Net n (  NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	    NetHandler* pN;

	    pN = n.connect(serverName, portNo);

	    CegoDbHandler* pSH = new CegoDbHandler(pN, pModule);

	    pModule->log(modId, Logger::DEBUG, Chain("Requesting session ..."));

	    CegoDbHandler::ResultType res = pSH->requestSession(tableSet, user, password);
	    if ( res != CegoDbHandler::DB_OK )
	    {
		Chain msg;
		pSH->getMsg(msg);
		throw Exception(EXLOC, msg);
	    }

	    long tid = pSH->getTid();

	    AbortHandler *pAH = new AbortHandler(serverName, portNo, tableSet, user, password, tid, pModule);
	    
	    pModule->log(modId, Logger::DEBUG, Chain("Got session ..."));
	    read_history((char*)histPath);
 
	    bool goOn = true;
	    while ( goOn )
	    {
		
		char *pC = 0;
		
		Chain cs;
		
		pC = readline(PROMPT);
		
		if (pC)
		{
		    cs = Chain(pC);

		    bool stmtComplete = false;
		    while (! stmtComplete )
		    {		
			add_history(pC);	
			
			if (pC)
			    free(pC);	    
			
			cs = cs.cutTrailing(Chain(" \t"));
			
			if ( cs.length() > 1)
			{		
			    if ( cs[cs.length()-2] == ';')
				stmtComplete = true;
			}
			
			if ( ! stmtComplete )
			{
			    if ( cs.length() > 1 )
			    {
				pC = readline(PROMPT2);
			    }
			    else
			    {
				pC = readline(PROMPT);
			    }
			    if (pC)
				cs = cs + Chain(" ") + Chain(pC);	   
			}
			
		    }
		    
		    Timer t(6,3);
		    t.start();

		    goOn = dispatchCmd(pSH, pAH, cs, rawMode, exitCode);

		    if ( exitCode != 0 )
			goOn = true;
				
		    // sending command to server here
		    
		    t.stop();
		    cerr << "ok ( " << t << " s )" << endl;
		    
		    write_history((char*)histPath);
		}
		else
		{
		    cout << "Bye" << endl;
		    goOn=false;
		}
	    }
	    
	    pSH->closeSession();

	    delete pSH;
	    delete pN;
	    delete pAH;
	    
	}
	catch ( Exception e)
	{
	    Chain msg;
	    e.pop(msg);
	    cerr << msg << endl;
	    exitCode=1;	   
	}
	break;
    }    
    }

    delete pModule;
    exit(exitCode);
}

bool dispatchCmd(CegoDbHandler *pSH, AbortHandler *pAH, Chain cmd, bool outMode, int& exitCode)
{
    
    exitCode = 0;

    if ( cmd.cutTrailing(" ;") == Chain("quit") )
	return false;

    Tokenizer tok(cmd, Chain(" ;"));
    Chain flowTok;
    if ( tok.nextToken(flowTok) )
    {
	
	if ( flowTok == Chain("flowctl") )
	{
	    Chain flowOpt;
	    if ( tok.nextToken(flowOpt) )
	    {
		if ( flowOpt == Chain("off") )
		{
		    flowCtl=false;
		    cout << "Flow control disabled" << endl;
		}
		else
		{
		    flowCtl=true;
		    if ( flowOpt.asInteger() > 0 )
			flowCount=flowOpt.asInteger();
		    cout << "Flow control enabled" << endl;    
		}
	    }	    
	    return true;
	}
    }
    
    Chain msg;
    CegoDbHandler::ResultType res;

    if ( pAH )
	pAH->setQueryAbort(true);
    
    res = pSH->reqQueryOp(cmd);
    
    if ( pAH )
	pAH->setQueryAbort(false);


    if ( res == CegoDbHandler::DB_OK )
    {
	Chain msg;
	pSH->getMsg(msg);
	long affCount = pSH->getAffected();

	// if ( msg == Chain("Goodbye") )
	//     return true;

	cout << msg << endl; // " ( " << affCount << " affected )" << endl;
	
    }
    else if ( res == CegoDbHandler::DB_INFO )
    {

	CegoObject::ObjectType objType = pSH->getObjType();

	switch ( objType )
	{
	case CegoObject::SYSTEM:
	case CegoObject::TABLE:
	case CegoObject::PINDEX:
	case CegoObject::UINDEX:
	case CegoObject::INDEX:
	{   
	    CegoTableObject to;
	    to.putElement( pSH->getObjElement() );
	    cout << to.getFormatted();
	    break;
	}
	case CegoObject::VIEW:
	{
	    CegoViewObject vo;
	    vo.putElement( pSH->getObjElement() );
	    cout << vo.getFormatted();
	    break;
	}
	case CegoObject::PROCEDURE:
	{
	    CegoProcObject po;
	    po.putElement( pSH->getObjElement() );
	    cout << po.getFormatted();
	    break;
	}
	case CegoObject::FKEY:
	{
	    CegoKeyObject ko;
	    ko.putElement( pSH->getObjElement() );
	    cout << ko.getFormatted();
	    break;
	}
	}
    }
    else if ( res == CegoDbHandler::DB_DATA )
    {
	ListT<CegoField> schema;
	Chain format;

	pSH->getSchema(schema);
	pSH->getFormat(format);

	CegoOutput output(schema, format);
	
	output.setRawMode(outMode);

	output.headOut();
	
	// CegoDbHandler::QuerySync querySync;

	// querySync=CegoDbHandler::QUERY_MOREDATA;

	ListT<CegoField> fvl;
	long count=0;
	
	while ( ( res = pSH->receiveTableData(schema, fvl) ) == CegoDbHandler::DB_DATA )
	{	    
	    count++;
	    
	    output.rowOut(fvl);
	    fvl.Empty();		    

	    if ( flowCtl )
	    {
		if ( count % flowCount == 0 )
		{
		    cout << "Next, Reset, Abort ? ( n, r, a ) ";
		    char resp;
		    cin >> resp;
		    if ( resp == 'a' )
		    {
			pSH->abortQuery();
			break;
		    }
		    if ( resp == 'r' )
		    {
			pSH->resetQuery();
		    }
		    else
		    {
			// nothing
		    }		    
		}
	    }
		
	}
	output.tailOut();

    }
    if ( res == CegoDbHandler::DB_ERROR )
    {
	Chain msg;
	pSH->getMsg(msg);
	cout << "Error : " << msg << endl;
	cout << "Query failed" << endl;
	exitCode = 1;
	return false;
	
    }

    return true;
}

AbortHandler::AbortHandler(const Chain& serverName, 
			   int portNo, 
			   const Chain& tableSet,
			   const Chain& user, 
			   const Chain& password,
			   long tid, 
			   CegoModule *pModule) :  SigHandler() 
{
    init();
    install(SIGINT);
#ifndef HAVE_MINGW32
    install(SIGPIPE);
#endif
    _tid = tid;
    _pModule = pModule;
    _portNo = portNo;
    _tableSet = tableSet;
    _user = user;
    _password = password;
    _serverName = serverName;
}

AbortHandler::~AbortHandler()
{
} 

void AbortHandler::sigCatch(int sig) 
{
    
    try {
	install(SIGINT);
#ifndef HAVE_MINGW32
	install(SIGPIPE);
#endif

	if ( _doAbort )
	{

	    cout << "Aborting query .." << endl;

	    Net n (  NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	    NetHandler* pN;
	    
	    pN = n.connect(_serverName, _portNo);
	    
	    CegoDbHandler* pSH = new CegoDbHandler(pN, _pModule);
	    	    
	    CegoDbHandler::ResultType res = pSH->requestSession(_tableSet, _user, _password);
	    if ( res != CegoDbHandler::DB_OK )
	    {
		Chain msg;
		pSH->getMsg(msg);
		throw Exception(EXLOC, msg);
	    }
    
	    Chain msg;
	    	    
	    res = pSH->reqQueryAbort(_tid);
	    	    
	    if ( res == CegoDbHandler::DB_OK )
	    {
		Chain msg;
		pSH->getMsg(msg);
	    }

	    
	    pSH->closeSession();

	    delete pSH;
	    delete pN;

	}
	else
	{
	    cout << "Use quit to quit" << endl;
	}
    }
    catch ( Exception e)
    {
	Chain msg;
	e.pop(msg);
	cout << "Error : " << msg << endl;
    }
    
}

void AbortHandler::setQueryAbort(bool doAbort)
{
    _doAbort = doAbort;
}
