///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoGroupCursor.cc
// -------------------
// Cego order cursor class implementation
//      
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-20010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoGroupCursor
// 
// Description: A cursor, to trace group space tables
//
// Status: QG-2.5.5
//
///////////////////////////////////////////////////////////////////////////////


#include "CegoGroupCursor.h"
#include "CegoGroupEntry.h"

CegoGroupCursor::CegoGroupCursor(int tabSetId, const Chain& groupSpace, CegoTableManager *pTM)
{
    _pTM = pTM;
    _groupSpace = groupSpace;
    _tabSetId = tabSetId;
}

CegoGroupCursor::~CegoGroupCursor()
{
    abort();
}

bool CegoGroupCursor::getFirst(ListT<CegoField>& fl)
{

    CegoObjectCursor* pC = _pTM->getObjectCursor(_tabSetId, _groupSpace, _groupSpace, CegoObject::ORDERSPACE);

    if ( ! pC )
    {
	Chain msg = "Cannot get cursor for <" + _groupSpace  + ">";
	throw Exception(EXLOC, msg);	
    } 
    
    char* p;
    int len;

    p = (char*)pC->getFirst(len, _curDP);

    delete pC;
    
    // root datapointer must not be zero
    if ( p == 0)
    {
	return false;
	// throw Exception(EXLOC, "Missing Group Anchor");
    }

    _bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, _curDP, p, len);    
    _curEntry.setPtr(p, len);

    CegoDataPointer nil;
    CegoDataPointer dp = _curEntry.getLeft();

    if ( dp == nil )
    {
	
	decodeFL(fl);
	
	
	return true;
    }
    else
    {
	return getMostLeft(fl);
    }

}

bool CegoGroupCursor::getMostLeft(ListT<CegoField>& fl)
{

    CegoDataPointer nil;

    while ( _curEntry.getLeft() != nil )
    {
	
	_curDP = _curEntry.getLeft();

	_pTM->releaseDataPtrUnlocked(_bp, false);
	
	char* p;
	int len;
	
	_bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, _curDP, p, len);
	
	_curEntry.setPtr(p, len);
    }

    decodeFL(fl);

    return true;
    
}

bool CegoGroupCursor::getNext(ListT<CegoField>& fl)
{

    CegoDataPointer nil;

    if ( _curEntry.getRight() == nil )
    {

	CegoDataPointer dp = _curDP;
	_curDP = _curEntry.getParent();
	
	_pTM->releaseDataPtrUnlocked(_bp, false);

	if ( _curDP == nil )
	    return false;
		
	char* p;
	int len;
	
	_bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, _curDP, p, len);
	_curEntry.setPtr(p, len);

	while ( _curEntry.getRight() == dp )
	{
	    	    
	    dp = _curDP;
	    _curDP = _curEntry.getParent();
	    
	    _pTM->releaseDataPtrUnlocked(_bp, false);
    
	    if ( _curDP == nil )
		return false;

	    _bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, _curDP, p, len);

	    _curEntry.setPtr(p, len);
	}
	
	decodeFL(fl);
		
	return true;
    }
    else
    {
	
	_curDP = _curEntry.getRight();	
	_pTM->releaseDataPtrUnlocked(_bp, false);
		
	char* p;
	int len;
	
	_bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, _curDP, p, len);
	
	_curEntry.setPtr(p, len);
	
	return getMostLeft(fl);	
    }
}

void CegoGroupCursor::decodeFL(ListT<CegoField>& fl)
{
    int tid;
    int tastep;
    CegoTupleState ts;
    
    _qh.decodeFVL(fl, (char*)_curEntry.getKeyPtr(), _curEntry.getKeyLen(), tid, tastep, ts, false);
    _qh.decodeFVL(fl, (char*)_curEntry.getTupPtr(), _curEntry.getTupLen(), tid, tastep, ts, false);	
    
    CegoField *pF = fl.First();
    while ( pF )
    {
	if ( pF->getAttrName() == Chain(AVG_GROUP) )
	{
	    // get count as last entry
	    CegoFieldValue fv = pF->getValue();
	    CegoFieldValue countValue = fl[fl.Size() - 1 ].getValue();
	    fv = fv / countValue;
	    pF->setValue(fv);
	}
	pF = fl.Next();
    }
}

void CegoGroupCursor::abort()
{
    if ( _bp.isFixed() )
	_pTM->releaseDataPtrUnlocked(_bp, false);
}
