///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoGroupSpace.cc
// ------------------
// Cego group space implementation
//     
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoGroupSpace
// 
// Description: 
//
// Status: QG-2.5.5
//
///////////////////////////////////////////////////////////////////////////////

#include "CegoGroupSpace.h"

#include "Exception.h"
#include "CegoDataType.h"
#include "CegoTableManager.h"
#include "CegoGroupEntry.h"
#include "CegoTypeConverter.h"

CegoGroupSpace::CegoGroupSpace(CegoTableManager *pTM)
{
    _pTM = pTM;
    _isCreated = false;
    _pCountAgg = 0;
}

CegoGroupSpace::~CegoGroupSpace()
{
    if ( _pCountAgg )
	delete _pCountAgg;
}

void CegoGroupSpace::createGroupSpace(int tabSetId, const Chain& groupSpace, ListT<CegoAttrDesc> *pGroupList, ListT<CegoAggregation*>& aggList)
{
    _tabSetId = tabSetId;
    _groupSpace = groupSpace;    
    _aggList = aggList;
    
    ListT<CegoField> groupSchema;
    int id=1;
    
    CegoAttrDesc* pAttrRef = pGroupList->First();
    while ( pAttrRef )
    {
	CegoField f(pAttrRef->getTableName(), pAttrRef->getAttrName());
	f.setId(id);
	groupSchema.Insert(f);
	id++;
	pAttrRef = pGroupList->Next();
    }

    // add count aggregation for average calculation
    bool hasAvgAgg = false;
    bool hasCountAgg = false;
    CegoAggregation **pAgg = _aggList.First();
    while ( pAgg )
    {
	if ( (*pAgg)->getType() == CegoAggregation::AVG )
	{
	    hasAvgAgg = true;
	}
	if ( (*pAgg)->getType() == CegoAggregation::COUNT )
	    hasCountAgg = true;
	pAgg = _aggList.Next();
    }
    
    if ( hasAvgAgg == true && hasCountAgg == false )
    {
	_pCountAgg = new CegoAggregation();
	_aggList.Insert( _pCountAgg );
    }
    
    pAgg = _aggList.First();
    while ( pAgg )
    {
	CegoField f;
	f.setId(id);

	if ( (*pAgg)->getType() == CegoAggregation::MIN)
	    f.setAttrName(MIN_GROUP);
	else if ( (*pAgg)->getType() == CegoAggregation::MAX)
	    f.setAttrName(MAX_GROUP);
	else if ( (*pAgg)->getType() == CegoAggregation::SUM)
	    f.setAttrName(SUM_GROUP);
	else if ( (*pAgg)->getType() == CegoAggregation::AVG)
	    f.setAttrName(AVG_GROUP);	
	else // COUNT
	    f.setAttrName(COUNT_GROUP);

	groupSchema.Insert(f);
	_aggSchema.Insert(f);
	(*pAgg)->setAggregationId(id);
	id++;
	pAgg = _aggList.Next();
    }

    _groupingOffset = pGroupList->Size() + 1;

    CegoTableObject oe;
    oe = CegoTableObject( _tabSetId, CegoObject::ORDERSPACE, _groupSpace, groupSchema, groupSpace);
    _pTM->createTableObject(oe);

    _pTM->getObject(_tabSetId, _groupSpace, CegoObject::ORDERSPACE, _oe, _obp);    

    _sysEntry = CegoDataPointer(_obp.getFileId(), _obp.getPageId(), _obp.getEntryPos());

    _pTM->getDBMng()->bufferUnfix(_obp, false, _pTM->getLockHandler());

     _isFirst = true;
     _isCreated = true;

}

void CegoGroupSpace::dropGroupSpace()
{    

    if ( _isCreated )
    {
	if ( _obp.isFixed() )
	    _pTM->getDBMng()->bufferUnfix(_obp, false, _pTM->getLockHandler());
	
	if ( _pTM->objectExists(_tabSetId, _groupSpace, CegoObject::ORDERSPACE) )
	    _pTM->removeObject(_tabSetId, _groupSpace, CegoObject::ORDERSPACE);
    }
    _isCreated = false;
}


ListT<CegoField> CegoGroupSpace::getSchema()
{
    return _oe.getSchema();    
}

void CegoGroupSpace::insertTuple(ListT<CegoField>& addKey, ListT<CegoField>& dataTuple)
{
    
    if ( _isCreated == false )
	throw Exception(EXLOC, "Groupspace not created, cannot insert tuple");

    int id=1;
    CegoField* pF = addKey.First();
    while (pF)
    {
	pF->setId(id);
	id++;
	pF = addKey.Next();
    }
    
    if ( _isFirst )
    {

	_isFirst = false;
	
	// enocode order list and fvl here
	
	char *keyPtr;
	int keyLen;

	CegoTupleState ts = COMMITTED;
	
	_qh.encodeFVL(0, 0, ts, addKey, keyPtr, keyLen);
	
	char *tupPtr;
	int tupLen;

	ListT<CegoField> groupedValues = initGrouping(dataTuple);

	_qh.encodeFVL(0, 0, ts, groupedValues, tupPtr, tupLen);
	
	CegoDataPointer nil;
	CegoGroupEntry noe;
	noe.initEntry(keyPtr, keyLen, tupPtr, tupLen, nil, nil, nil);

	try
	{
	    _rdp = _pTM->insertData(_sysEntry, _oe, (char*)noe.getPtr(), noe.getLen(), true);	    
	}
	catch ( Exception e )
	{
	    throw Exception(EXLOC, "Cannot insert order entry", e);
	}

	free(keyPtr);	    
	free(tupPtr);

    }
    else
    {
	insertTupleInTree(addKey, dataTuple);
    }
}

CegoGroupCursor* CegoGroupSpace::getCursor()
{
    if ( _isCreated == false )
	throw Exception(EXLOC, "Groupspace not created, cannot insert tuple");

    CegoGroupCursor *pOC = new CegoGroupCursor(_tabSetId, _groupSpace, _pTM);
    return pOC;
}


void CegoGroupSpace::insertTupleInTree(ListT<CegoField>& addKey, ListT<CegoField>& dataTuple)
{


    ListT<CegoField> actKey = _oe.getSchema();

    char* p;
    int len;

    int tid = 0;
    int tastep = 0;

    CegoDataPointer pdp = _rdp;
    
    // unsigned long lockId;
    CegoBufferPage bp;

    // lockId = _pTM->claimGroupDataPtr(_tabSetId, CegoLockHandler::WRITE, CegoBufferPool::SYNC, pdp, p, len);
    bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, pdp, p, len);
    
    CegoDataPointer nil;
    CegoDataPointer ndp;

    CegoGroupEntry actEntry;
    actEntry.setPtr(p, len);

    CegoTupleState ts;    
  
    _qh.decodeFVL(actKey, (char*)actEntry.getKeyPtr(), actEntry.getKeyLen(), tid, tastep, ts);
        
    while ( 1 ) 
    {

	if  ( compKey(actKey, addKey) == HIGHER ) 
	{
	    
	    if (  actEntry.getLeft() == nil )
	    {

		// insert at left
				
		char *keyPtr;
		int keyLen;
		_qh.encodeFVL(tid, tastep, ts, addKey, keyPtr, keyLen);

		ListT<CegoField> groupedValues = initGrouping(dataTuple);
		
		char *tupPtr;
		int tupLen;
		_qh.encodeFVL(tid, tastep, ts, groupedValues, tupPtr, tupLen);
		
		CegoGroupEntry newEntry;
		newEntry.initEntry((char*)keyPtr, keyLen, (char*)tupPtr, tupLen, nil, nil, pdp);

		free(keyPtr);
		free(tupPtr);
				
		CegoDataPointer idp;
		idp = _pTM->insertData(_sysEntry, _oe, (char*)newEntry.getPtr(), newEntry.getLen(), true);
		
		actEntry.setLeft(idp);				
		_pTM->releaseDataPtrUnlocked(bp, true);
		
		return;
		
	    }
	    else
	    {
		pdp = actEntry.getLeft();

		_pTM->releaseDataPtrUnlocked(bp, false);

		bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, pdp, p, len);
		
		actEntry.setPtr(p, len);
		_qh.decodeFVL(actKey, (char*)actEntry.getKeyPtr(), actEntry.getKeyLen(), tid, tastep, ts);
	    }
	}
	else if  ( compKey(actKey, addKey) == LOWER ) 
	{
	    
	    if ( actEntry.getRight() == nil )
	    {

		// insert at right
		
		char *keyPtr;
		int keyLen;
		_qh.encodeFVL(tid, tastep, ts, addKey, keyPtr, keyLen);

		ListT<CegoField> groupedValues = initGrouping(dataTuple);
		
		char *tupPtr;
		int tupLen;

		_qh.encodeFVL(tid, tastep, ts, groupedValues, tupPtr, tupLen);
		
		CegoGroupEntry newEntry;
		newEntry.initEntry((char*)keyPtr, keyLen, (char*)tupPtr, tupLen, nil, nil, pdp);
		
		free(keyPtr);
		free(tupPtr);
		
		CegoDataPointer idp;
		idp = _pTM->insertData(_sysEntry, _oe, (char*)newEntry.getPtr(), newEntry.getLen(), true);
				
		actEntry.setRight(idp);

		_pTM->releaseDataPtrUnlocked(bp, true);

		return;
		
	    }
	    else
	    {
		pdp = actEntry.getRight();

		_pTM->releaseDataPtrUnlocked(bp, false);		

		bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, pdp, p, len);

		actEntry.setPtr(p, len);
		_qh.decodeFVL(actKey, (char*)actEntry.getKeyPtr(), actEntry.getKeyLen(), tid, tastep, ts);
	    }	    
	}
	else
	{
	    // make grouping

	   
	    ListT<CegoField> groupValue = _aggSchema;

	    _qh.decodeFVL(groupValue, (char*)actEntry.getTupPtr(), actEntry.getTupLen(), tid, tastep, ts);

	    ListT<CegoField> newGroupValue = aggGrouping(groupValue, dataTuple);

	    char *tupPtr;
	    int tupLen;

	    _qh.encodeFVL(tid, tastep, ts, newGroupValue, tupPtr, tupLen);
	    memcpy(actEntry.getTupPtr(), tupPtr, actEntry.getTupLen());
	    free(tupPtr);
		
	    _pTM->releaseDataPtrUnlocked(bp, true);

	    return;	    

	}
    }
}

CegoGroupSpace::KeyComp CegoGroupSpace::compKey(ListT<CegoField>& key1, ListT<CegoField>& key2)
{
    CegoField *pF1 = key1.First();
    CegoField *pF2 = key2.First();

    while ( pF1 && pF2 )
    {

	if ( pF1->getValue() < pF2->getValue() )
	{
	    return LOWER;
	}
	else if ( pF1->getValue() > pF2->getValue() )
	{
	    return HIGHER;
	}

	pF1 = key1.Next();
	pF2 = key2.Next();
	
    }

    return EQUAL;
}


ListT<CegoField> CegoGroupSpace::initGrouping(ListT<CegoField>& dataTuple)
{

    ListT<CegoField> initValue;

    CegoAggregation** pAgg = _aggList.First();
    int i=_groupingOffset;
    while ( pAgg )
    {
	CegoField f;
	f.setId(i);
        
	switch ( (*pAgg)->getType() )
	{	    
	case CegoAggregation::COUNT:
	{
	    f.setValue(CegoFieldValue(INT_TYPE, Chain("1")));
	    break;
	}	
	case CegoAggregation::SUM:
	case CegoAggregation::AVG:
	case CegoAggregation::MIN:
	case CegoAggregation::MAX:
	{	    
	    f.setValue( getValueForAgg(*pAgg, dataTuple));
	    break;
	}	
	}

	initValue.Insert(f);
	
	i++;
	pAgg = _aggList.Next();
    }
    
    return initValue;
}

ListT<CegoField> CegoGroupSpace::aggGrouping(ListT<CegoField>& aggValues, ListT<CegoField>& dataTuple)
{
    
    CegoAggregation** pAgg = _aggList.First();
    CegoField *pF = aggValues.First();

    while ( pAgg && pF )
    {
	(*pAgg)->setFieldValue(pF->getValue());
	pAgg = _aggList.Next();
	pF = aggValues.Next();
    }

    CegoQueryHelper qh;    

    ListT<CegoField> aggValue;

    int i=0;
    pAgg = _aggList.First();
    while ( pAgg  )
    {
	qh.aggregateTuple(dataTuple, *pAgg );
	
	CegoField f;
	f.setAttrName( _aggSchema[i].getAttrName());
	f.setValue( (*pAgg)->getFieldValue() );
	f.setId(i + _groupingOffset);
	i++;
	aggValue.Insert(f);
	pAgg = _aggList.Next();
    }

    return aggValue;
    
}

CegoFieldValue CegoGroupSpace::getValueForAgg(CegoAggregation* pAgg, ListT<CegoField>& dataTuple)
{
	
    CegoField f;
    CegoFieldValue fv;

    CegoFactor *pAF = pAgg->getFactor();
    
    if ( pAF == 0 )
    {
	f.setAttrName(Chain("*"));
	fv.setType(INT_TYPE);
    }
    else
    {
	switch ( pAF->getType() )
	{	    
	case CegoFactor::ATTR:
	{
	    if (CegoField* pF = dataTuple.Find(CegoField(pAF->getAttr().getTableName(), pAF->getAttr().getAttrName())))
	    {
		fv = pF->getValue();
		f = *pF;
	    }
	    break;
	}
	case CegoFactor::CONSTVAL:
	{	
	    fv = pAF->getFieldValue();
	    f = CegoField(pAF->getFieldValue()); 
	    break;
	}
	case CegoFactor::EXPR:
	{
	    pAF->setFieldListArray( &dataTuple, 1 );
	    fv = pAF->evalFieldValue();
	    break;
	}
	}
	
	if ( fv.getValue() == 0 && pAgg->getType() != CegoAggregation::MIN )
	{
	    switch ( fv.getType() )
	    {
	    case INT_TYPE:
	    {			
		fv = CegoFieldValue(INT_TYPE, Chain("0"));
		break;
	    }
	    case VARCHAR_TYPE:
	    {
		fv = CegoFieldValue(VARCHAR_TYPE, Chain(""));
		break;
	    }
	    case DECIMAL_TYPE:
	    {
		fv = CegoFieldValue(DECIMAL_TYPE, Chain("0.0"));
		break;		   
	    }
	    case FIXED_TYPE:
	    {
		fv = CegoFieldValue(FIXED_TYPE, Chain("0.0"));
		break;		   
	    }
	    default:
	    {
		Chain msg = Chain("Aggregation not supported on datatype <") + CEGO_TYPE_MAP[ (int) fv.getType()] + Chain(">");
		throw Exception(EXLOC, msg);
		break;
	    }
	    }
	}
    }
    
    return fv;
}



